import datetime
import os
import re
import sys
import time
import unicodedata
import warnings
from base64 import b64encode
from decimal import Decimal
from logging import Logger
from pathlib import Path
from typing import Any, Dict, List, Union

import pytz
import requests
from requests import Response

from erepublik import __version__, constants

try:
    import simplejson as json
except ImportError:
    import json

__all__ = [
    "VERSION",
    "calculate_hit",
    "date_from_eday",
    "eday_from_date",
    "deprecation",
    "get_final_hit_dmg",
    "write_file",
    "get_air_hit_dmg_value",
    "get_file",
    "get_ground_hit_dmg_value",
    "get_sleep_seconds",
    "good_timedelta",
    "slugify",
    "interactive_sleep",
    "json",
    "localize_dt",
    "localize_timestamp",
    "normalize_html_json",
    "now",
    "silent_sleep",
    "json_decode_object_hook",
    "json_load",
    "json_loads",
    "json_dump",
    "json_dumps",
    "b64json",
    "ErepublikJSONEncoder",
]

VERSION: str = __version__


def now() -> datetime.datetime:
    return datetime.datetime.now(constants.erep_tz).replace(microsecond=0)


def localize_timestamp(timestamp: int) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(timestamp, constants.erep_tz)


def localize_dt(dt: Union[datetime.date, datetime.datetime]) -> datetime.datetime:
    if isinstance(dt, datetime.datetime):
        return constants.erep_tz.localize(dt)
    elif isinstance(dt, datetime.date):
        return constants.erep_tz.localize(datetime.datetime.combine(dt, datetime.time(0, 0, 0)))
    else:
        raise TypeError(f"Argument dt must be and instance of datetime.datetime or datetime.date not {type(dt)}")


def good_timedelta(dt: datetime.datetime, td: datetime.timedelta) -> datetime.datetime:
    """Normalize timezone aware datetime object after timedelta to correct jumps over DST switches

    :param dt: Timezone aware datetime object
    :type dt: datetime.datetime
    :param td: timedelta object
    :type td: datetime.timedelta
    :return: datetime object with correct timezone when jumped over DST
    :rtype: datetime.datetime
    """
    return constants.erep_tz.normalize(dt + td)


def eday_from_date(date: Union[datetime.date, datetime.datetime] = None) -> int:
    if date is None:
        date = now()
    if isinstance(date, datetime.date):
        date = datetime.datetime.combine(date, datetime.time(0, 0, 0))
    return (date - datetime.datetime(2007, 11, 20, 0, 0, 0)).days


def date_from_eday(eday: int) -> datetime.date:
    return localize_dt(datetime.date(2007, 11, 20)) + datetime.timedelta(days=eday)


def get_sleep_seconds(time_until: datetime.datetime) -> int:
    """time_until aware datetime object Wrapper for sleeping until"""
    sleep_seconds = int((time_until - now()).total_seconds())
    return sleep_seconds if sleep_seconds > 0 else 0


def interactive_sleep(sleep_seconds: int):
    while sleep_seconds > 0:
        seconds = sleep_seconds
        if (seconds - 1) // 1800:
            seconds = seconds % 1800 if seconds % 1800 else 1800
        elif (seconds - 1) // 300:
            seconds = seconds % 300 if seconds % 300 else 300
        elif (seconds - 1) // 60:
            seconds = seconds % 60 if seconds % 60 else 60
        # elif (seconds - 1) // 30:
        #     seconds = seconds % 30 if seconds % 30 else 30
        else:
            seconds = 1
        sys.stdout.write(f"\rSleeping for {sleep_seconds:4} more seconds")
        sys.stdout.flush()
        time.sleep(seconds)
        sleep_seconds -= seconds
    sys.stdout.write("\r")


silent_sleep = time.sleep


def get_file(filepath: str) -> str:
    file = Path(filepath)
    if file.exists():
        if file.is_dir():
            return str(file / "new_file.txt")
        else:
            version = 1
            try:
                version = int(file.suffix[1:]) + 1
                basename = file.stem
            except ValueError:
                basename = file.name
                version += 1

            full_name = file.parent / f"{basename}.{version}"
            while full_name.exists():
                version += 1
                full_name = file.parent / f"{basename}.{version}"
            return str(full_name)
    else:
        os.makedirs(file.parent, exist_ok=True)
        return str(file)


def write_file(filename: str, content: str) -> int:
    filename = get_file(filename)
    with open(filename, "ab") as f:
        ret = f.write(content.encode("utf-8"))
    return ret


def normalize_html_json(js: str) -> str:
    js = re.sub(r" \'(.*?)\'", lambda a: f'"{a.group(1)}"', js)
    js = re.sub(r"(\d\d):(\d\d):(\d\d)", r"\1\2\3", js)
    js = re.sub(r'([{\s,])(\w+)(:)(?!"})', r'\1"\2"\3', js)
    js = re.sub(r",\s*}", "}", js)
    return js


def slugify(value, allow_unicode=False) -> str:
    """
    Function copied from Django2.2.1 django.utils.text.slugify
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^\w\s-]", "_", value).strip().lower()
    return re.sub(r"[-\s]+", "-", value)


def calculate_hit(
    strength: float, rang: int, tp: bool, elite: bool, ne: bool, booster: int = 0, weapon: int = 200, is_deploy: bool = False
) -> Decimal:
    dec = 3 if is_deploy else 0
    base_str = 1 + Decimal(str(round(strength, 3))) / 400
    base_rnk = 1 + Decimal(str(rang / 5))
    base_wpn = 1 + Decimal(str(weapon / 100))
    dmg = 10 * base_str * base_rnk * base_wpn

    dmg = get_final_hit_dmg(dmg, rang, tp=tp, elite=elite, ne=ne, booster=booster)
    return Decimal(round(dmg, dec))


def get_ground_hit_dmg_value(
    citizen_id: int, natural_enemy: bool = False, true_patriot: bool = False, booster: int = 0, weapon_power: int = 200
) -> Decimal:
    r = requests.get(f"https://www.erepublik.com/en/main/citizen-profile-json/{citizen_id}").json()
    rang = r["military"]["militaryData"]["ground"]["rankNumber"]
    strength = r["military"]["militaryData"]["ground"]["strength"]
    elite = r["citizenAttributes"]["level"] > 100
    if natural_enemy:
        true_patriot = True

    return calculate_hit(strength, rang, true_patriot, elite, natural_enemy, booster, weapon_power)


def get_air_hit_dmg_value(
    citizen_id: int, natural_enemy: bool = False, true_patriot: bool = False, booster: int = 0, weapon_power: int = 0
) -> Decimal:
    r = requests.get(f"https://www.erepublik.com/en/main/citizen-profile-json/{citizen_id}").json()
    rang = r["military"]["militaryData"]["aircraft"]["rankNumber"]
    elite = r["citizenAttributes"]["level"] > 100
    return calculate_hit(0, rang, true_patriot, elite, natural_enemy, booster, weapon_power)


def get_final_hit_dmg(
    base_dmg: Union[Decimal, float, str], rang: int, tp: bool = False, elite: bool = False, ne: bool = False, booster: int = 0
) -> Decimal:
    dmg = Decimal(str(base_dmg))

    if elite:
        dmg = dmg * 11 / 10
    if tp and rang >= 70:
        dmg = dmg * (1 + Decimal((rang - 69) / 10))
    dmg = dmg * (100 + booster) / 100
    if ne:
        dmg = dmg * 11 / 10
    return Decimal(dmg)


def deprecation(message):
    warnings.warn(message, DeprecationWarning, stacklevel=2)


def json_decode_object_hook(
    o: Union[Dict[str, Any], List[Any], int, float, str]
) -> Union[Dict[str, Any], List[Any], int, float, str, datetime.date, datetime.datetime, datetime.timedelta]:
    """Convert classes.ErepublikJSONEncoder datetime, date and timedelta to their python objects

    :param o:
    :return: Union[Dict[str, Any], List[Any], int, float, str, datetime.date, datetime.datetime, datetime.timedelta]
    """
    if o.get("__type__"):
        _type = o.get("__type__")
        if _type == "datetime":
            dt = datetime.datetime.strptime(f"{o['date']} {o['time']}", "%Y-%m-%d %H:%M:%S")
            if o.get("tzinfo"):
                dt = pytz.timezone(o["tzinfo"]).localize(dt)
            return dt
        elif _type == "date":
            dt = datetime.datetime.strptime(f"{o['date']}", "%Y-%m-%d")
            return dt.date()
        elif _type == "timedelta":
            return datetime.timedelta(seconds=o["total_seconds"])
    return o


def json_load(f, **kwargs):
    # kwargs.update(object_hook=json_decode_object_hook)
    return json.load(f, **kwargs)


def json_loads(s: str, **kwargs):
    # kwargs.update(object_hook=json_decode_object_hook)
    return json.loads(s, **kwargs)


def json_dump(obj, fp, *args, **kwargs):
    if not kwargs.get("cls"):
        kwargs.update(cls=ErepublikJSONEncoder)
    return json.dump(obj, fp, *args, **kwargs)


def json_dumps(obj, *args, **kwargs):
    if not kwargs.get("cls"):
        kwargs.update(cls=ErepublikJSONEncoder)
    return json.dumps(obj, *args, **kwargs)


def b64json(obj: Union[Dict[str, Union[int, List[str]]], List[str]]):
    if isinstance(obj, list):
        return b64encode(json.dumps(obj, separators=(",", ":")).encode("utf-8")).decode("utf-8")
    elif isinstance(obj, (int, str)):
        return obj
    elif isinstance(obj, dict):
        for k, v in obj.items():
            obj[k] = b64json(v)
    else:
        from .classes import ErepublikException

        raise ErepublikException(f"Unhandled object type! obj is {type(obj)}")
    return b64encode(json.dumps(obj, separators=(",", ":")).encode("utf-8")).decode("utf-8")


class ErepublikJSONEncoder(json.JSONEncoder):
    def default(self, o):
        try:
            from erepublik.citizen import Citizen

            if isinstance(o, Decimal):
                return float(f"{o:.02f}")
            elif isinstance(o, datetime.datetime):
                return dict(
                    __type__="datetime",
                    date=o.strftime("%Y-%m-%d"),
                    time=o.strftime("%H:%M:%S"),
                    tzinfo=str(o.tzinfo) if o.tzinfo else None,
                )
            elif isinstance(o, datetime.date):
                return dict(__type__="date", date=o.strftime("%Y-%m-%d"))
            elif isinstance(o, datetime.timedelta):
                return dict(
                    __type__="timedelta", days=o.days, seconds=o.seconds, microseconds=o.microseconds, total_seconds=o.total_seconds()
                )
            elif isinstance(o, Response):
                return dict(headers=dict(o.__dict__["headers"]), url=o.url, text=o.text, status_code=o.status_code)
            elif hasattr(o, "as_dict"):
                return o.as_dict
            elif isinstance(o, set):
                return list(o)
            elif isinstance(o, Citizen):
                return o.to_json()
            elif isinstance(o, Logger):
                return str(o)
            elif hasattr(o, "__dict__"):
                return o.__dict__
            else:
                return super().default(o)
        except Exception as e:  # noqa
            return str(e)



import datetime
import hashlib
import threading
import warnings
import weakref
from decimal import Decimal
from io import BytesIO
from typing import Any, Dict, Generator, Iterable, List, NamedTuple, NoReturn, Optional, Tuple, Union

from requests import HTTPError, Response, Session, post

from erepublik import _types as types
from erepublik import constants, utils

__all__ = [
    "Battle",
    "BattleDivision",
    "BattleSide",
    "Company",
    "Config",
    "Details",
    "Energy",
    "ErepublikException",
    "ErepublikNetworkException",
    "EnergyToFight",
    "Holding",
    "Inventory",
    "MyCompanies",
    "OfferItem",
    "Politics",
    "Reporter",
    "TelegramReporter",
]


class ErepublikException(Exception):
    def __init__(self, message):
        super().__init__(message)


class ErepublikNetworkException(ErepublikException):
    def __init__(self, message, request):
        super().__init__(message)
        self.request = request


class CloudFlareSessionError(ErepublikNetworkException):
    pass


class CaptchaSessionError(ErepublikNetworkException):
    pass


class Holding:
    id: int
    region: int
    companies: List["Company"]
    name: str
    _citizen = weakref.ReferenceType

    def __init__(self, _id: int, region: int, citizen, name: str = None):
        self._citizen = weakref.ref(citizen)
        self.id: int = _id
        self.region: int = region
        self.companies: List["Company"] = list()
        if name:
            self.name = name
        else:
            comp_sum = len(self.companies)
            name = f"Holding (#{self.id}) with {comp_sum} "
            if comp_sum == 1:
                name += "company"
            else:
                name += "companies"
            self.name = name

    @property
    def wam_count(self) -> int:
        return len([1 for company in self.companies if company.wam_enabled and not company.already_worked])

    @property
    def wam_companies(self) -> Iterable["Company"]:
        return [company for company in self.companies if company.wam_enabled]

    @property
    def employable_companies(self) -> Iterable["Company"]:
        return [company for company in self.companies if company.preset_works]

    def add_company(self, company: "Company") -> NoReturn:
        self.companies.append(company)
        self.companies.sort()

    def get_wam_raw_usage(self) -> Dict[str, Decimal]:
        frm = Decimal("0.00")
        wrm = Decimal("0.00")
        for company in self.wam_companies:
            if company.industry in [1, 7, 8, 9, 10, 11]:
                frm += company.raw_usage
            elif company.industry in [2, 12, 13, 14, 15, 16]:
                wrm += company.raw_usage
        return dict(frm=frm, wrm=wrm)

    def get_wam_companies(self, raw_factory: bool = None) -> List["Company"]:
        raw = []
        factory = []
        for company in self.wam_companies:
            if not company.already_worked and not company.cannot_wam_reason == "war":
                if company.is_raw:
                    raw.append(company)
                else:
                    factory.append(company)
        if raw_factory is None:
            return raw + factory
        else:
            return raw if raw_factory else factory

    def __str__(self) -> str:
        comp = len(self.companies)
        name = f"Holding (#{self.id}) with {comp} "
        if comp == 1:
            name += "company"
        else:
            name += "companies"
        return name

    def __repr__(self):
        return str(self)

    @property
    def as_dict(self) -> Dict[str, Union[str, int, List[Dict[str, Union[str, int, bool, float, Decimal]]]]]:
        return dict(name=self.name, id=self.id, region=self.region, companies=[c.as_dict for c in self.companies], wam_count=self.wam_count)

    @property
    def citizen(self):
        return self._citizen()


class Company:
    _holding: weakref.ReferenceType
    holding: Holding
    id: int
    quality: int
    is_raw: bool
    raw_usage: Decimal
    products_made: Decimal
    wam_enabled: bool
    can_wam: bool
    cannot_wam_reason: str
    industry: int
    already_worked: bool
    preset_works: int

    def __init__(
        self,
        holding: Holding,
        _id: int,
        quality: int,
        is_raw: bool,
        effective_bonus: Decimal,
        raw_usage: Decimal,
        base_production: Decimal,
        wam_enabled: bool,
        can_wam: bool,
        cannot_wam_reason: str,
        industry: int,
        already_worked: bool,
        preset_works: int,
    ):
        self._holding = weakref.ref(holding)
        self.id: int = _id
        self.industry: int = industry
        self.quality: int = self._get_real_quality(quality)
        self.is_raw: bool = is_raw
        self.wam_enabled: bool = wam_enabled
        self.can_wam: bool = can_wam
        self.cannot_wam_reason: str = cannot_wam_reason
        self.already_worked: bool = already_worked
        self.preset_works: int = preset_works

        self.products_made = self.raw_usage = Decimal(base_production) * Decimal(effective_bonus)
        if not self.is_raw:
            self.raw_usage = -self.products_made * raw_usage

    def _get_real_quality(self, quality) -> int:
        #  7: 'FRM q1',  8: 'FRM q2',  9: 'FRM q3', 10: 'FRM q4', 11: 'FRM q5',
        # 12: 'WRM q1', 13: 'WRM q2', 14: 'WRM q3', 15: 'WRM q4', 16: 'WRM q5',
        # 18: 'HRM q1', 19: 'HRM q2', 20: 'HRM q3', 21: 'HRM q4', 22: 'HRM q5',
        # 24: 'ARM q1', 25: 'ARM q2', 26: 'ARM q3', 27: 'ARM q4', 28: 'ARM q5',
        if 7 <= self.industry <= 11:
            return self.industry % 6
        elif 12 <= self.industry <= 16:
            return self.industry % 11
        elif 18 <= self.industry <= 22:
            return self.industry % 17
        elif 24 <= self.industry <= 28:
            return self.industry % 23
        else:
            return quality

    @property
    def _internal_industry(self) -> int:
        #  7: 'FRM q1',  8: 'FRM q2',  9: 'FRM q3', 10: 'FRM q4', 11: 'FRM q5',
        # 12: 'WRM q1', 13: 'WRM q2', 14: 'WRM q3', 15: 'WRM q4', 16: 'WRM q5',
        # 18: 'HRM q1', 19: 'HRM q2', 20: 'HRM q3', 21: 'HRM q4', 22: 'HRM q5',
        # 24: 'ARM q1', 25: 'ARM q2', 26: 'ARM q3', 27: 'ARM q4', 28: 'ARM q5',
        if 7 <= self.industry <= 11:
            return 7
        elif 12 <= self.industry <= 16:
            return 12
        elif 18 <= self.industry <= 22:
            return 18
        elif 24 <= self.industry <= 28:
            return 24
        else:
            return self.industry

    @property
    def _sort_keys(self):
        return not self.is_raw, self._internal_industry, self.quality, self.id

    def __hash__(self):
        return hash(self._sort_keys)

    def __lt__(self, other: "Company"):
        return self._sort_keys < other._sort_keys

    def __le__(self, other: "Company"):
        return self._sort_keys <= other._sort_keys

    def __gt__(self, other: "Company"):
        return self._sort_keys > other._sort_keys

    def __ge__(self, other: "Company"):
        return self._sort_keys >= other._sort_keys

    def __eq__(self, other: "Company"):
        return self._sort_keys == other._sort_keys

    def __ne__(self, other: "Company"):
        return self._sort_keys != other._sort_keys

    def __str__(self):
        name = f"(#{self.id:>9d}) {constants.INDUSTRIES[self.industry]}"
        if not self.is_raw:
            name += f" q{self.quality}"
        return name

    def __repr__(self):
        return str(self)

    @property
    def as_dict(self) -> Dict[str, Union[str, int, bool, float, Decimal]]:
        return dict(
            name=str(self),
            holding=self.holding.id,
            id=self.id,
            quality=self.quality,
            is_raw=self.is_raw,
            raw_usage=self.raw_usage,
            products_made=self.products_made,
            wam_enabled=self.wam_enabled,
            can_wam=self.can_wam,
            cannot_wam_reason=self.cannot_wam_reason,
            industry=self.industry,
            already_worked=self.already_worked,
            preset_works=self.preset_works,
        )

    def dissolve(self) -> Response:
        self.holding.citizen.write_log(f"{self} dissolved!")
        # noinspection PyProtectedMember
        return self.holding.citizen._post_economy_sell_company(self.id, self.holding.citizen.details.pin, sell=False)

    def upgrade(self, level: int) -> Response:
        # noinspection PyProtectedMember
        return self.holding.citizen._post_economy_upgrade_company(self.id, level, self.holding.citizen.details.pin)

    @property
    def holding(self) -> Holding:
        return self._holding()


class MyCompanies:
    work_units: int = 0
    next_ot_time: datetime.datetime
    ff_lockdown: int = 0

    holdings: Dict[int, Holding]
    _companies: weakref.WeakSet
    _citizen: weakref.ReferenceType
    companies: Generator[Company, None, None]

    def __init__(self, citizen):
        self._citizen = weakref.ref(citizen)
        self.holdings = dict()
        self._companies = weakref.WeakSet()
        self.next_ot_time = utils.now()

    def prepare_holdings(self, holdings: Dict[str, Dict[str, Any]]):
        """
        :param holdings: Parsed JSON to dict from en/economy/myCompanies
        """
        for holding in holdings.values():
            if holding.get("id") not in self.holdings:
                self.holdings.update({int(holding.get("id")): Holding(holding["id"], holding["region_id"], self.citizen, holding["name"])})
        if not self.holdings.get(0):
            self.holdings.update({0: Holding(0, 0, self.citizen, "Unassigned")})  # unassigned

    def prepare_companies(self, companies: Dict[str, Dict[str, Any]]):
        """
        :param companies: Parsed JSON to dict from en/economy/myCompanies
        """
        self.__clear_data()
        for company_dict in companies.values():
            holding = self.holdings.get(int(company_dict["holding_company_id"]))
            quality = company_dict.get("quality")
            is_raw = company_dict.get("is_raw")
            if is_raw:
                raw_usage = Decimal("0.0")
            else:
                raw_usage = Decimal(str(company_dict.get("upgrades").get(str(quality)).get("raw_usage")))
            company = Company(
                holding,
                company_dict.get("id"),
                quality,
                is_raw,
                Decimal(str(company_dict.get("effective_bonus"))) / 100,
                raw_usage,
                Decimal(str(company_dict.get("base_production"))),
                company_dict.get("wam_enabled"),
                company_dict.get("can_work_as_manager"),
                company_dict.get("cannot_work_as_manager_reason"),
                company_dict.get("industry_id"),
                company_dict.get("already_worked"),
                company_dict.get("preset_works"),
            )
            self._companies.add(company)
            holding.add_company(company)

    def get_employable_factories(self) -> Dict[int, int]:
        return {company.id: company.preset_works for company in self.companies if company.preset_works}

    def get_total_wam_count(self) -> int:
        return sum([holding.wam_count for holding in self.holdings.values()])

    @staticmethod
    def get_needed_inventory_usage(companies: Union[Company, Iterable[Company]]) -> Decimal:
        if isinstance(companies, list):
            return sum(company.products_made * 100 if company.is_raw else 1 for company in companies)
        else:
            return companies.products_made

    @property
    def companies(self) -> Generator[Company, None, None]:
        return (c for c in self._companies)

    def __str__(self):
        return f"MyCompanies: {sum(1 for _ in self.companies)} companies in {len(self.holdings)} holdings"

    def __repr__(self):
        return str(self)

    def __clear_data(self):
        for holding in self.holdings.values():
            for company in holding.companies:  # noqa
                del company
            holding.companies.clear()
        self._companies.clear()

    @property
    def as_dict(
        self,
    ) -> Dict[
        str,
        Union[str, int, datetime.datetime, Dict[str, Dict[str, Union[str, int, List[Dict[str, Union[str, int, bool, float, Decimal]]]]]]],
    ]:
        return dict(
            name=str(self),
            work_units=self.work_units,
            next_ot_time=self.next_ot_time,
            ff_lockdown=self.ff_lockdown,
            holdings={str(hi): h.as_dict for hi, h in self.holdings.items()},
            company_count=sum(1 for _ in self.companies),
        )

    @property
    def citizen(self):
        return self._citizen()


class Config:
    email = ""
    password = ""
    work = True
    train = True
    wam = False
    ot = True
    auto_sell: List[str] = None
    auto_sell_all = False
    employees = False
    fight = False
    air = False
    ground = False
    all_in = False
    next_energy = False
    boosters = False
    travel_to_fight = False
    always_travel = False
    epic_hunt = False
    epic_hunt_ebs = False
    rw_def_side = False
    interactive = True
    continuous_fighting = False
    auto_buy_raw = False
    force_wam = False
    sort_battles_time = True
    force_travel = False
    telegram = True
    telegram_chat_id = 0
    telegram_token = ""
    maverick = False
    spin_wheel_of_fortune = False

    def __init__(self):
        self.auto_sell = []

    def reset(self):
        self.work = True
        self.train = True
        self.wam = False
        self.ot = True
        self.auto_sell = list()
        self.auto_sell_all = False
        self.employees = False
        self.fight = False
        self.air = False
        self.ground = False
        self.all_in = False
        self.next_energy = False
        self.boosters = False
        self.travel_to_fight = False
        self.always_travel = False
        self.epic_hunt = False
        self.epic_hunt_ebs = False
        self.rw_def_side = False
        self.interactive = True
        self.continuous_fighting = False
        self.auto_buy_raw = False
        self.force_wam = False
        self.sort_battles_time = True
        self.force_travel = False
        self.telegram = True
        self.telegram_chat_id = 0
        self.telegram_token = ""
        self.maverick = False
        self.spin_wheel_of_fortune = False

    @property
    def as_dict(self) -> Dict[str, Union[bool, int, str, List[str]]]:
        return dict(
            email=self.email,
            work=self.work,
            train=self.train,
            wam=self.wam,
            ot=self.ot,
            auto_sell=self.auto_sell,
            auto_sell_all=self.auto_sell_all,
            employees=self.employees,
            fight=self.fight,
            air=self.air,
            ground=self.ground,
            all_in=self.all_in,
            next_energy=self.next_energy,
            travel_to_fight=self.travel_to_fight,
            always_travel=self.always_travel,
            epic_hunt=self.epic_hunt,
            epic_hunt_ebs=self.epic_hunt_ebs,
            rw_def_side=self.rw_def_side,
            interactive=self.interactive,
            maverick=self.maverick,
            continuous_fighting=self.continuous_fighting,
            auto_buy_raw=self.auto_buy_raw,
            force_wam=self.force_wam,
            sort_battles_time=self.sort_battles_time,
            force_travel=self.force_travel,
            telegram=self.telegram,
            telegram_chat_id=self.telegram_chat_id,
            telegram_token=self.telegram_token,
            spin_wheel_of_fortune=self.spin_wheel_of_fortune,
        )


class Energy:
    limit = 500  # energyToRecover
    interval = 10  # energyPerInterval
    energy = 0  # energy
    _recovery_time = None

    def __init__(self):
        self._recovery_time = utils.now()

    def __repr__(self):
        return f"{self.energy:4}/{self.limit:4}, {self.interval:3}hp/6min"

    @property
    def recovered(self):
        warnings.warn("Deprecated since auto auto-eat! Will be removed soon. Use Energy.energy", DeprecationWarning)
        return self.energy

    @property
    def recoverable(self):
        warnings.warn("Deprecated since auto auto-eat! Will be removed soon. Use Energy.energy", DeprecationWarning)
        return 0

    def set_reference_time(self, recovery_time: datetime.datetime):
        self._recovery_time = recovery_time.replace(microsecond=0)

    @property
    def food_fights(self):
        return self.energy // 10

    @property
    def reference_time(self):
        if self.is_energy_full or self._recovery_time < utils.now():
            ret = utils.now()
        else:
            ret = self._recovery_time
        return ret

    @property
    def is_recoverable_full(self):
        warnings.warn("Deprecated since auto auto-eat! Will be removed soon. Use Energy.is_energy_full", DeprecationWarning)
        return self.is_energy_full

    @property
    def is_recovered_full(self):
        warnings.warn("Deprecated since auto auto-eat! Will be removed soon. Use Energy.is_energy_full", DeprecationWarning)
        return self.is_energy_full

    @property
    def is_energy_full(self):
        return self.energy >= self.limit - self.interval

    @property
    def available(self):
        warnings.warn("Deprecated since auto auto-eat! Will be removed soon. Use Energy.energy", DeprecationWarning)
        return self.energy

    @property
    def as_dict(self) -> Dict[str, Union[int, datetime.datetime, bool]]:
        return dict(
            limit=self.limit,
            interval=self.interval,
            energy=self.energy,
            reference_time=self.reference_time,
            food_fights=self.food_fights,
            is_energy_full=self.is_energy_full,
        )


class Details:
    xp: int = 0
    cc: float = 0
    pp: int = 0
    pin: str = None
    gold: float = 0
    level: int = 0
    next_pp: List[int] = None
    citizen_id: int = 0
    citizenship: constants.Country
    current_region: int = 0
    current_country: constants.Country
    residence_region: int = 0
    residence_country: constants.Country
    daily_task_done: bool = False
    daily_task_reward: bool = False
    mayhem_skills: Dict[int, int]

    def __init__(self):
        self.next_pp = []
        self.mayhem_skills = {1: 0, 2: 0, 3: 0, 4: 0, 5: 0, 6: 0, 7: 0, 8: 0, 9: 0, 10: 0, 11: 0, 12: 0, 13: 0, 14: 0}
        _default_country = constants.Country(0, "Unknown", "Unknown", "XX")
        self.citizenship = self.current_country = self.residence_country = _default_country

    @property
    def xp_till_level_up(self):
        if self.xp >= 10000:
            next_level_up = (1 + (self.xp // 5000)) * 5000
        elif self.xp >= 7000:
            next_level_up = 10000
        elif self.xp >= 3000:
            next_level_up = (1 + ((self.xp - 1000) // 2000)) * 2000 + 1000
        elif self.xp >= 2000:
            next_level_up = 3000
        elif self.xp >= 450:
            next_level_up = (1 + (self.xp // 500)) * 500
        elif self.xp >= 370:
            next_level_up = (1 + ((self.xp - 10) // 40)) * 40 + 10
        elif self.xp >= 300:
            next_level_up = (1 + ((self.xp - 20) // 35)) * 35 + 20
        elif self.xp >= 150:
            next_level_up = (1 + (self.xp // 30)) * 30
        elif self.xp >= 50:
            next_level_up = (1 + ((self.xp - 10) // 20)) * 20 + 10
        elif self.xp >= 20:
            next_level_up = (1 + ((self.xp - 5) // 15)) * 15 + 5
        else:
            next_level_up = (1 + (self.xp // 10)) * 10
        return next_level_up - self.xp

    @property
    def as_dict(self) -> Dict[str, Union[int, float, str, constants.Country, bool]]:
        return dict(
            xp=self.xp,
            cc=self.cc,
            pp=self.pp,
            pin=self.pin,
            gold=self.gold,
            next_pp=self.next_pp,
            level=self.level,
            citizen_id=self.citizen_id,
            citizenship=self.citizenship,
            current_region=self.current_region,
            current_country=self.current_country,
            residence_region=self.residence_region,
            residence_country=self.residence_country,
            daily_task_done=self.daily_task_done,
            daily_task_reward=self.daily_task_reward,
            mayhem_skills=self.mayhem_skills,
            xp_till_level_up=self.xp_till_level_up,
        )

    @property
    def is_elite(self):
        return self.level > 100


class Politics:
    is_party_member: bool = False

    party_id: int = 0
    party_slug: str = ""
    is_party_president: bool = False
    is_congressman: bool = False
    is_country_president: bool = False

    @property
    def as_dict(self) -> Dict[str, Union[bool, int, str]]:
        return dict(
            is_party_member=self.is_party_member,
            party_id=self.party_id,
            party_slug=self.party_slug,
            is_party_president=self.is_party_president,
            is_congressman=self.is_congressman,
            is_country_president=self.is_country_president,
        )


class House:
    quality = None
    unactivated_count = 0
    active_until = utils.good_timedelta(utils.now(), -datetime.timedelta(days=1))

    def __init__(self, quality: int):
        if 0 < quality < 6:
            self.quality = quality

    @property
    def next_ot_point(self) -> datetime.datetime:
        return self.active_until


class Reporter:
    __to_update: List[Dict[Any, Any]] = None
    key: str = ""
    allowed: bool = False

    @property
    def name(self) -> str:
        return self.citizen.name

    @property
    def email(self) -> str:
        return self.citizen.config.email

    @property
    def citizen_id(self) -> int:
        return self.citizen.details.citizen_id

    @property
    def as_dict(self) -> Dict[str, Union[bool, int, str, List[Dict[Any, Any]]]]:
        return dict(
            name=self.name, email=self.email, citizen_id=self.citizen_id, key=self.key, allowed=self.allowed, queue=self.__to_update
        )

    def __init__(self, citizen):
        self._citizen = weakref.ref(citizen)
        self._req = Session()
        self.url = "https://api.erep.lv"
        self._req.headers.update(
            {
                "user-agent": "eRepublik Script Reporter v3",
                "erep-version": utils.__version__,
                "erep-user-id": str(self.citizen_id),
                "erep-user-name": self.citizen.name,
            }
        )
        self.__to_update = []
        self.__registered: bool = False

    def do_init(self):
        self.key: str = ""
        self.__update_key()
        self.register_account()
        self.allowed = True

    @property
    def citizen(self):
        return self._citizen()

    def __update_key(self):
        self.key = hashlib.md5(bytes(f"{self.name}:{self.email}", encoding="UTF-8")).hexdigest()

    def __bot_update(self, data: dict) -> Response:
        if self.__to_update:
            for unreported_data in self.__to_update:
                unreported_data.update(player_id=self.citizen_id, key=self.key)
                unreported_data = utils.json_loads(utils.json_dumps(unreported_data))
                r = self._req.post(f"{self.url}/bot/update", json=unreported_data)
                r.raise_for_status()
            self.__to_update.clear()
        data = utils.json.loads(utils.json_dumps(data))
        r = self._req.post(f"{self.url}/bot/update", json=data)
        r.raise_for_status()
        return r

    def _bot_update(self, data: Dict[str, Any]) -> Optional[Response]:
        if not self.__registered:
            self.do_init()
        if self.allowed:
            try:
                return self.__bot_update(data)
            except HTTPError:
                self.__to_update.append(data)
        else:
            self.__to_update.append(data)

    def register_account(self):
        if not self.__registered:
            r = self.__bot_update(dict(key=self.key, check=True, player_id=self.citizen_id))
            if r:
                if not r.json().get("status"):
                    self._req.post(f"{self.url}/bot/register", json=dict(name=self.name, email=self.email, player_id=self.citizen_id))
                self.__registered = True
                self.allowed = True
                self.report_action("STARTED", value=utils.now().strftime("%F %T"))

    def send_state_update(
        self,
        xp: int,
        cc: float,
        gold: float,
        inv_total: int,
        inv: int,
        hp_limit: int,
        hp_interval: int,
        hp_available: int,
        food: int,
        pp: int,
    ):

        data = dict(
            key=self.key,
            player_id=self.citizen_id,
            state=dict(
                xp=xp,
                cc=cc,
                gold=gold,
                inv_total=inv_total,
                inv_free=inv_total - inv,
                inv=inv,
                food=food,
                pp=pp,
                hp_limit=hp_limit,
                hp_interval=hp_interval,
                hp_available=hp_available,
            ),
        )
        self._bot_update(data)

    def report_action(self, action: str, json_val: Dict[Any, Any] = None, value: str = None):
        json_data = dict(player_id=getattr(self, "citizen_id", None), log={"action": action}, key=getattr(self, "key", None))
        if json_val:
            json_data["log"].update(dict(json=json_val))
        if value:
            json_data["log"].update(dict(value=value))
        if not any([self.key, self.email, self.name, self.citizen_id]):
            return
        self._bot_update(json_data)

    def report_fighting(self, battle: "Battle", invader: bool, division: "BattleDivision", damage: float, hits: int):
        side = battle.invader if invader else battle.defender
        self.report_action(
            "FIGHT",
            dict(
                battle_id=battle.id,
                side=side,
                dmg=damage,
                air=battle.has_air,
                hits=hits,
                round=battle.zone_id,
                extra=dict(battle=battle, side=side, division=division),
            ),
        )

    def report_money_donation(self, citizen_id: int, amount: float, is_currency: bool = True):
        cur = "cc" if is_currency else "gold"
        self.report_action(
            "DONATE_MONEY",
            dict(citizen_id=citizen_id, amount=amount, currency=cur),
            f"Successfully donated {amount}{cur} to citizen with id {citizen_id}!",
        )

    def report_item_donation(self, citizen_id: int, amount: float, quality: int, industry: str):
        self.report_action(
            "DONATE_ITEMS",
            dict(citizen_id=citizen_id, amount=amount, quality=quality, industry=industry),
            f"Successfully donated {amount} x {industry} q{quality} to citizen with id {citizen_id}!",
        )

    def report_promo(self, kind: str, time_until: datetime.datetime):
        self._req.post(f"{self.url}/promos/add/", data=dict(kind=kind, time_untill=time_until))

    def fetch_battle_priorities(self, country: constants.Country) -> List["Battle"]:
        try:
            battle_response = self._req.get(f"{self.url}/api/v1/battles/{country.id}")
            return [
                self.citizen.all_battles[bid] for bid in battle_response.json().get("battle_ids", []) if bid in self.citizen.all_battles
            ]
        except:  # noqa
            return []

    def fetch_tasks(self) -> List[Dict[str, Any]]:
        try:
            task_response = self._req.post(f"{self.url}/api/v1/command", data=dict(citizen=self.citizen_id, key=self.key)).json()
            if task_response.get("status"):
                return task_response.get("data")
            else:
                return []
        except:  # noqa
            return []


class BattleSide:
    points: int
    deployed: List[constants.Country]
    allies: List[constants.Country]
    battle: "Battle"
    _battle: weakref.ReferenceType
    country: constants.Country
    is_defender: bool

    def __init__(
        self,
        battle: "Battle",
        country: constants.Country,
        points: int,
        allies: List[constants.Country],
        deployed: List[constants.Country],
        defender: bool,
    ):
        self._battle = weakref.ref(battle)
        self.country = country
        self.points = points
        self.allies = allies
        self.deployed = deployed
        self.is_defender = defender

    @property
    def id(self) -> int:
        return self.country.id

    def __repr__(self):
        side_text = "Defender" if self.is_defender else "Invader "
        return f"<BattleSide: {side_text} {self.country.name}|{self.points:>2d}p>"

    def __str__(self):
        side_text = "Defender" if self.is_defender else "Invader "
        return f"{side_text} {self.country.name} - {self.points:>2d} points"

    def __format__(self, format_spec):
        return self.country.iso

    @property
    def as_dict(self) -> Dict[str, Union[int, constants.Country, bool, List[constants.Country]]]:
        return dict(points=self.points, country=self.country, is_defender=self.is_defender, allies=self.allies, deployed=self.deployed)

    @property
    def battle(self):
        return self._battle()


class BattleDivision:
    id: int
    end: datetime.datetime
    epic: bool
    dom_pts: Dict[str, int]
    wall: Dict[str, Union[int, float]]
    def_medal: Dict[str, int]
    inv_medal: Dict[str, int]
    terrain: int
    div: int
    battle: "Battle"
    _battle: weakref.ReferenceType

    @property
    def as_dict(self):
        return dict(
            id=self.id, division=self.div, terrain=(self.terrain, self.terrain_display), wall=self.wall, epic=self.epic, end=self.div_end
        )

    @property
    def is_air(self):
        return self.div == 11

    @property
    def div_end(self) -> bool:
        return utils.now() >= self.end

    def __init__(
        self,
        battle: "Battle",
        div_id: int,
        end: datetime.datetime,
        epic: bool,
        div: int,
        wall_for: int,
        wall_dom: float,
        terrain_id: int = 0,
    ):
        """Battle division helper class

        :type div_id: int
        :type end: datetime.datetime
        :type epic: bool
        :type div: int
        :type terrain_id: int
        :type wall_for: int
        :type wall_dom: float
        """
        self._battle = weakref.ref(battle)
        self.id = div_id
        self.end = end
        self.epic = epic
        self.wall = {"for": wall_for, "dom": wall_dom}
        self.terrain = terrain_id
        self.div = div

    @property
    def terrain_display(self):
        return constants.TERRAINS[self.terrain]

    def __str__(self):
        base_name = f"D{self.div} #{self.id}"
        if self.terrain:
            base_name += f" ({self.terrain_display})"
        if self.div_end:
            base_name += " Ended"
        return base_name

    def __repr__(self):
        return f"<BattleDivision #{self.id} (battle #{self.battle.id})>"

    @property
    def battle(self):
        return self._battle()


class Battle:
    id: int
    war_id: int
    zone_id: int
    is_rw: bool
    is_dict_lib: bool
    start: datetime.datetime
    invader: BattleSide
    defender: BattleSide
    div: Dict[int, BattleDivision]
    region_id: int
    region_name: str

    @property
    def as_dict(self):
        return dict(
            id=self.id,
            war_id=self.war_id,
            divisions=self.div,
            zone=self.zone_id,
            rw=self.is_rw,
            dict_lib=self.is_dict_lib,
            start=self.start,
            sides={"inv": self.invader, "def": self.defender},
            region=[self.region_id, self.region_name],
            link=self.link,
        )

    @property
    def has_air(self) -> bool:
        for div in self.div.values():
            if div.div == 11:
                return True
        return not bool(self.zone_id % 4)

    @property
    def has_started(self) -> bool:
        return self.start <= utils.now()

    @property
    def has_ground(self) -> bool:
        for div in self.div.values():
            if div.div != 11:
                return True
        return bool(self.zone_id % 4)

    @property
    def link(self):
        return f"https://www.erepublik.com/en/military/battlefield/{self.id}"

    def __init__(self, battle: Dict[str, Any]):
        """Object representing eRepublik battle.

        :param battle: Dict object for single battle from '/military/campaignsJson/list' response's 'battles' object
        """

        self.id = int(battle.get("id"))
        self.war_id = int(battle.get("war_id"))
        self.zone_id = int(battle.get("zone_id"))
        self.is_rw = bool(battle.get("is_rw"))
        self.is_as = bool(battle.get("is_as"))
        self.is_dict_lib = bool(battle.get("is_dict")) or bool(battle.get("is_lib"))
        self.region_id = battle.get("region", {}).get("id")
        self.region_name = battle.get("region", {}).get("name")
        self.start = datetime.datetime.fromtimestamp(int(battle.get("start", 0)), tz=constants.erep_tz)

        self.invader = BattleSide(
            self,
            constants.COUNTRIES[battle.get("inv", {}).get("id")],
            battle.get("inv", {}).get("points"),
            [constants.COUNTRIES[row.get("id")] for row in battle.get("inv", {}).get("ally_list")],
            [constants.COUNTRIES[row.get("id")] for row in battle.get("inv", {}).get("ally_list") if row["deployed"]],
            False,
        )

        self.defender = BattleSide(
            self,
            constants.COUNTRIES[battle.get("def", {}).get("id")],
            battle.get("def", {}).get("points"),
            [constants.COUNTRIES[row.get("id")] for row in battle.get("def", {}).get("ally_list")],
            [constants.COUNTRIES[row.get("id")] for row in battle.get("def", {}).get("ally_list") if row["deployed"]],
            True,
        )

        self.div = {}
        for div, data in battle.get("div", {}).items():
            div = int(div)
            if data.get("end"):
                end = datetime.datetime.fromtimestamp(data.get("end"), tz=constants.erep_tz)
            else:
                end = constants.max_datetime

            battle_div = BattleDivision(
                self,
                div_id=data.get("id"),
                div=data.get("div"),
                end=end,
                epic=data.get("epic_type") in [1, 5],
                wall_for=data.get("wall").get("for"),
                wall_dom=data.get("wall").get("dom"),
                terrain_id=data.get("terrain", 0),
            )

            self.div.update({div: battle_div})

    def __str__(self):
        time_now = utils.now()
        is_started = self.start < utils.now()
        if is_started:
            time_part = f" {time_now - self.start}"
        else:
            time_part = f"-{self.start - time_now}"

        return (
            f"Battle {self.id} for {self.region_name[:16]:16} | "
            f"{self.invader} : {self.defender} | Round time {time_part} | {'R'+str(self.zone_id):>3}"
        )

    def __repr__(self):
        return f"<Battle #{self.id} {self.invader}:{self.defender} R{self.zone_id}>"


class EnergyToFight:
    energy: int = 0

    def __init__(self, energy: int = 0):
        self.energy = energy

    def __int__(self):
        return self.energy

    def __str__(self):
        return str(self.energy)

    def __repr__(self):
        return str(self.energy)

    @property
    def i(self):
        return self.__int__()

    @property
    def s(self):
        return self.__str__()

    def check(self, new_energy: int):
        if not isinstance(new_energy, (tuple, int)):
            return self.energy
        if 0 < new_energy < self.energy:
            self.energy = new_energy
        return self.energy


class TelegramReporter:
    __initialized: bool = False
    __queue: List[str]
    chat_id: int = 0
    api_url: str = ""
    player_name: str = ""
    __thread_stopper: threading.Event
    _last_time: datetime.datetime
    _last_full_energy_report: datetime.datetime
    _next_time: datetime.datetime
    _threads: List[threading.Thread]

    def __init__(self, stop_event: threading.Event = None):
        self._threads = []
        self.__queue = []
        self.__thread_stopper = threading.Event() if stop_event is None else stop_event
        self._last_full_energy_report = self._last_time = utils.good_timedelta(utils.now(), datetime.timedelta(hours=1))
        self._next_time = utils.now()

    @property
    def as_dict(self):
        return {
            "chat_id": self.chat_id,
            "api_url": self.api_url,
            "player": self.player_name,
            "last_time": self._last_time,
            "next_time": self._next_time,
            "queue": self.__queue,
            "initialized": self.__initialized,
            "has_threads": not self._threads,
        }

    def do_init(self, chat_id: int, token: str = None, player_name: str = None):
        if token is None:
            token = "864251270:AAFzZZdjspI-kIgJVk4gF3TViGFoHnf8H4o"
        self.chat_id = chat_id
        self.api_url = f"https://api.telegram.org/bot{token}"
        self.player_name = player_name or ""
        self.__initialized = True
        self._last_time = utils.good_timedelta(utils.now(), datetime.timedelta(minutes=-5))
        self._last_full_energy_report = utils.good_timedelta(utils.now(), datetime.timedelta(minutes=-30))
        if self.__queue:
            self.send_message("Telegram initialized")

    def send_message(self, message: str) -> bool:
        self.__queue.append(message)
        if not self.__initialized:
            if self._last_time < utils.now():
                self.__queue.clear()
            return True
        self._threads = [t for t in self._threads if t.is_alive()]
        self._next_time = utils.good_timedelta(utils.now(), datetime.timedelta(seconds=20))
        if not self._threads:
            name = f"telegram_{f'{self.player_name}_' if self.player_name else ''}send"
            send_thread = threading.Thread(target=self.__send_messages, name=name)
            send_thread.start()
            self._threads.append(send_thread)

        return True

    def report_full_energy(self, available: int, limit: int, interval: int):
        if (utils.now() - self._last_full_energy_report).total_seconds() >= 30 * 60:
            self._last_full_energy_report = utils.now()
            message = f"Full energy ({available}hp/{limit}hp +{interval}hp/6min)"
            self.send_message(message)

    def report_medal(self, msg, multiple: bool = True):
        new_line = "\n" if multiple else ""
        self.send_message(f"New award: {new_line}*{msg}*")

    def report_fight(self, battle: "Battle", invader: bool, division: "BattleDivision", damage: float, hits: int):
        side_txt = (battle.invader if invader else battle.defender).country.iso
        self.send_message(
            f"*Fight report*:\n{int(damage):,d} dmg ({hits} hits) in"
            f" [battle {battle.id} for {battle.region_name[:16]}]({battle.link}) in d{division.div} on "
            f"{side_txt} side"
        )

    def report_item_donation(self, citizen_id: int, amount: float, product: str):
        self.send_message(
            f"*Donation*: {amount} x {product} to citizen " f"[{citizen_id}](https://www.erepublik.com/en/citizen/profile/{citizen_id})"
        )

    def report_money_donation(self, citizen_id: int, amount: float, is_currency: bool = True):
        self.send_message(
            f"*Donation*: {amount}{'cc' if is_currency else 'gold'} to citizen "
            f"[{citizen_id}](https://www.erepublik.com/en/citizen/profile/{citizen_id})"
        )

    def __send_messages(self):
        while self._next_time > utils.now():
            if self.__thread_stopper.is_set():
                break
            self.__thread_stopper.wait(utils.get_sleep_seconds(self._next_time))

        message = "\n\n".join(self.__queue)
        if self.player_name:
            message = f"Player *{self.player_name}*\n\n" + message
        response = post(f"{self.api_url}/sendMessage", json=dict(chat_id=self.chat_id, text=message, parse_mode="Markdown"))
        self._last_time = utils.now()
        if response.json().get("ok"):
            self.__queue.clear()
            return True
        return False

    def send_photos(self, photos: List[Tuple[str, BytesIO]]):
        for photo_title, photo in photos:
            photo.seek(0)
            post(
                f"https://{self.api_url}/sendPhoto",
                data=dict(chat_id=self.chat_id, caption=photo_title),
                files=[("photo", ("f{utils.slugify(photo_title)}.png", photo))],
            )
        return


class OfferItem(NamedTuple):
    price: float = 999_999_999.0
    country: constants.Country = constants.Country(0, "", "", "")
    amount: int = 0
    offer_id: int = 0
    citizen_id: int = 0


class Inventory:
    final: types.InvFinal
    active: types.InvFinal
    boosters: types.InvBooster
    raw: types.InvRaw
    market: types.InvRaw
    used: int
    total: int

    def __init__(self):
        self.active = {}
        self.final = {}
        self.boosters = {}
        self.raw = {}
        self.offers = {}
        self.used = 0
        self.total = 0

    @property
    def as_dict(self) -> Dict[str, Union[types.InvFinal, types.InvRaw, int]]:
        return dict(
            active=self.active, final=self.final, boosters=self.boosters, raw=self.raw, offers=self.offers, total=self.total, used=self.used
        )



import base64
import datetime
import inspect
import logging
import os
import sys
import weakref
from logging import LogRecord, handlers
from pathlib import Path
from typing import Any, Dict, Union

import requests

from erepublik.classes import Reporter
from erepublik.constants import erep_tz
from erepublik.utils import json, json_dumps, json_loads, slugify


class ErepublikFileHandler(handlers.TimedRotatingFileHandler):
    _file_path: Path

    def __init__(self, filename: str = "log/erepublik.log", *args, **kwargs):
        log_path = Path(filename)
        self._file_path = log_path
        log_path.parent.mkdir(parents=True, exist_ok=True)
        at_time = erep_tz.localize(datetime.datetime.now()).replace(hour=0, minute=0, second=0, microsecond=0)
        kwargs.update(atTime=at_time)
        super().__init__(filename, when="d", *args, **kwargs)

    def doRollover(self) -> None:
        self._file_path.parent.mkdir(parents=True, exist_ok=True)
        super().doRollover()

    def emit(self, record: LogRecord) -> None:
        self._file_path.parent.mkdir(parents=True, exist_ok=True)
        super().emit(record)


class ErepublikLogConsoleHandler(logging.StreamHandler):
    def __init__(self, *_):
        super().__init__(sys.stdout)


class ErepublikFormatter(logging.Formatter):
    """override logging.Formatter to use an aware datetime object"""

    dbg_fmt = "[%(asctime)s] DEBUG: %(module)s: %(lineno)d: %(msg)s"
    info_fmt = "[%(asctime)s] %(msg)s"
    default_fmt = "[%(asctime)s] %(levelname)s: %(msg)s"

    def converter(self, timestamp: Union[int, float]) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(timestamp).astimezone(erep_tz)

    def format(self, record: logging.LogRecord) -> str:
        """
        Format the specified record as text.

        The record's attribute dictionary is used as the operand to a
        string formatting operation which yields the returned string.
        Before formatting the dictionary, a couple of preparatory steps
        are carried out. The message attribute of the record is computed
        using LogRecord.getMessage(). If the formatting string uses the
        time (as determined by a call to usesTime(), formatTime() is
        called to format the event time. If there is exception information,
        it is formatted using formatException() and appended to the message.
        """
        if record.levelno == logging.DEBUG:
            self._fmt = self.dbg_fmt
        elif record.levelno == logging.INFO:
            self._fmt = self.info_fmt
        else:
            self._fmt = self.default_fmt
        self._style = logging.PercentStyle(self._fmt)

        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        s = self.formatMessage(record)
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        return s

    def formatTime(self, record, datefmt=None):
        dt = self.converter(record.created)
        if datefmt:
            s = dt.strftime(datefmt)
        else:
            s = dt.strftime("%Y-%m-%d %H:%M:%S")
        return s

    def usesTime(self):
        return self._style.usesTime()


class ErepublikErrorHTTTPHandler(handlers.HTTPHandler):
    def __init__(self, reporter: Reporter):
        logging.Handler.__init__(self, level=logging.ERROR)
        self._reporter = weakref.ref(reporter)
        self.host = "erep.lv"
        self.url = "/ebot/error/"
        self.method = "POST"
        self.secure = True
        self.credentials = (str(reporter.citizen_id), reporter.key)
        self.context = None

    @property
    def reporter(self):
        return self._reporter()

    def _get_last_response(self) -> Dict[str, str]:
        response = self.reporter.citizen.r
        url = response.url
        last_index = url.index("?") if "?" in url else len(response.url)

        name = slugify(response.url[len(self.reporter.citizen.url) : last_index])
        html = response.text

        try:
            json_loads(html)
            ext = "json"
        except json.decoder.JSONDecodeError:
            ext = "html"
        try:
            resp_time = (
                datetime.datetime.strptime(response.headers.get("date"), "%a, %d %b %Y %H:%M:%S %Z")
                .replace(tzinfo=datetime.timezone.utc)
                .astimezone(erep_tz)
                .strftime("%F_%H-%M-%S")
            )
        except:  # noqa
            resp_time = slugify(response.headers.get("date"))
        return dict(
            name=f"{resp_time}_{name}.{ext}", content=html.encode("utf-8"), mimetype="application/json" if ext == "json" else "text/html"
        )

    def _get_local_vars(self) -> str:
        trace = inspect.trace()
        local_vars = {}
        if trace:
            local_vars = trace[-1][0].f_locals
            if local_vars.get("__name__") == "__main__":
                local_vars.update(
                    commit_id=local_vars.get("COMMIT_ID"),
                    interactive=local_vars.get("INTERACTIVE"),
                    version=local_vars.get("__version__"),
                    config=local_vars.get("CONFIG"),
                )
        else:
            stack = inspect.stack()
            report_error_caller_found = False
            for frame in stack:
                if report_error_caller_found:
                    local_vars = frame.frame.f_locals
                    break
                if "report_error" in str(frame.frame):
                    report_error_caller_found = True

        if "state_thread" in local_vars:
            local_vars.pop("state_thread", None)
        from erepublik import Citizen

        if isinstance(local_vars.get("self"), Citizen):
            local_vars["self"] = repr(local_vars["self"])
        if isinstance(local_vars.get("player"), Citizen):
            local_vars["player"] = repr(local_vars["player"])
        if isinstance(local_vars.get("citizen"), Citizen):
            local_vars["citizen"] = repr(local_vars["citizen"])
        return json_dumps(local_vars)

    def _get_instance_json(self) -> str:
        if self.reporter:
            return self.reporter.citizen.to_json(False)
        return ""

    def mapLogRecord(self, record: logging.LogRecord) -> Dict[str, Any]:
        data = super().mapLogRecord(record)

        # Log last response
        resp = self._get_last_response()
        files = [
            ("file", (resp.get("name"), resp.get("content"), resp.get("mimetype"))),
        ]

        files += list(("file", (f, open(f"log/{f}", "rb"))) for f in os.listdir("log") if f.endswith(".log"))
        local_vars_json = self._get_local_vars()
        if local_vars_json:
            files.append(("file", ("local_vars.json", local_vars_json, "application/json")))
        instance_json = self._get_instance_json()
        if instance_json:
            files.append(("file", ("instance.json", instance_json, "application/json")))
        data.update(files=files)
        return data

    def emit(self, record):
        """
        Emit a record.

        Send the record to the Web server as a percent-encoded dictionary
        """
        try:
            proto = "https" if self.secure else "http"
            u, p = self.credentials
            s = "Basic " + base64.b64encode(f"{u}:{p}".encode("utf-8")).strip().decode("ascii")
            headers = {"Authorization": s}
            data = self.mapLogRecord(record)
            files = data.pop("files") if "files" in data else None
            requests.post(f"{proto}://{self.host}{self.url}", headers=headers, data=data, files=files)
        except Exception:
            self.handleError(record)



import datetime
import hashlib
import random
import time
from typing import Any, Dict, List, Mapping, Union

from requests import Response, Session
from requests_toolbelt.utils import dump

from erepublik import constants, utils

__all__ = ["SlowRequests", "CitizenAPI"]


class SlowRequests(Session):
    last_time: datetime.datetime
    timeout: datetime.timedelta = datetime.timedelta(milliseconds=500)
    debug: bool = False

    def __init__(self, proxies: Dict[str, str] = None, user_agent: str = None):
        super().__init__()
        if proxies:
            self.proxies = proxies
        if user_agent is None:
            user_agent = random.choice(self.get_random_user_agent())
        self.request_log_name = utils.get_file(utils.now().strftime("debug/requests_%Y-%m-%d.log"))
        self.last_time = utils.now()
        self.headers.update({"User-Agent": user_agent})
        self.hooks["response"] = [self._log_response]

    @property
    def as_dict(self):
        return dict(
            last_time=self.last_time,
            timeout=self.timeout,
            cookies=self.cookies.get_dict(),
            debug=self.debug,
            user_agent=self.headers["User-Agent"],
            request_log_name=self.request_log_name,
            proxies=self.proxies,
        )

    def request(self, method, url, *args, **kwargs):
        self._slow_down_requests()
        self._log_request(url, method, **kwargs)
        resp = super().request(method, url, *args, **kwargs)
        # self._log_response(resp)
        return resp

    def _slow_down_requests(self):
        ltt = utils.good_timedelta(self.last_time, self.timeout)
        if ltt > utils.now():
            seconds = (ltt - utils.now()).total_seconds()
            time.sleep(seconds if seconds > 0 else 0)
        self.last_time = utils.now()

    def _log_request(self, url, method, data=None, json=None, params=None, **kwargs):
        if self.debug:
            args = {}
            kwargs.pop("allow_redirects", None)
            if kwargs:
                args.update({"kwargs": kwargs})

            if data:
                args.update({"data": data})

            if json:
                args.update({"json": json})

            if params:
                args.update({"params": params})

            body = f"[{utils.now().strftime('%F %T')}]\tURL: '{url}'\tMETHOD: {method}\tARGS: {args}\n"
            with open(self.request_log_name, "ab") as file:
                file.write(body.encode("UTF-8"))

    def _log_response(self, response: Response, *args, **kwargs):
        redirect = kwargs.get("redirect")
        url = response.request.url
        if self.debug:
            if response.history and not redirect:
                for hist_resp in response.history:
                    self._log_request(hist_resp.request.url, "REDIRECT")
                    self._log_response(hist_resp, redirect=True)

            fd_path = "debug/requests"
            fd_time = self.last_time.strftime("%Y/%m/%d/%H-%M-%S")
            fd_name = utils.slugify(url[len(CitizenBaseAPI.url) :])
            fd_extra = "_REDIRECT" if redirect else ""

            try:
                utils.json.loads(response.text)
                fd_ext = "json"
            except utils.json.JSONDecodeError:
                fd_ext = "html"

            filename = f"{fd_path}/{fd_time}_{fd_name}{fd_extra}.{fd_ext}"
            utils.write_file(filename, response.text)

            if not redirect:
                data = dump.dump_all(response)
                utils.write_file(f"debug/dumps/{fd_time}_{fd_name}{fd_extra}.{fd_ext}.dump", data.decode("utf8"))

    @staticmethod
    def get_random_user_agent() -> str:
        windows_x64 = "Windows NT 10.0; Win64; x64"
        linux_x64 = "X11; Linux x86_64"
        android_11 = "Android 11; Mobile"
        android_10 = "Android 10; Mobile"
        android_9 = "Android 9; Mobile"

        firefox_tmplt = "Mozilla/5.0 ({osystem}; rv:{version}.0) Gecko/20100101 Firefox/{version}.0"
        ff_version = range(85, 92)

        chrome_tmplt = "Mozilla/5.0 ({osystem}) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{version} Safari/537.36"
        chrome_version = [
            "85.0.4183.121",
            "86.0.4240.183",
            "87.0.4280.141",
            "88.0.4324.182",
            "89.0.4389.128",
            "90.0.4430.18",
            "91.0.4472.73",
            "92.0.4515.14",
        ]
        uas = []

        for osystem in [windows_x64, linux_x64, android_9, android_10, android_11]:
            for version in ff_version:
                uas.append(firefox_tmplt.format(osystem=osystem, version=version))
            for version in chrome_version:
                uas.append(chrome_tmplt.format(osystem=osystem, version=version))

        return random.choice(uas)


class CitizenBaseAPI:
    url: str = "https://www.erepublik.com/en"
    _req: SlowRequests
    token: str

    def __init__(self):
        """Class for unifying eRepublik known endpoints and their required/optional parameters"""
        self._req = SlowRequests()
        self.token = ""

    @property
    def as_dict(self):
        return dict(url=self.url, request=self._req.as_dict, token=self.token)

    def post(self, url: str, data=None, json=None, **kwargs) -> Response:
        return self._req.post(url, data, json, **kwargs)

    def get(self, url: str, **kwargs) -> Response:
        return self._req.get(url, **kwargs)

    def _get_main(self) -> Response:
        return self.get(self.url)

    def set_socks_proxy(self, host: str, port: int, username: str = None, password: str = None):
        url = f"socks5://{username}:{password}@{host}:{port}" if username and password else f"socks5://{host}:{port}"
        self._req.proxies = dict(http=url, https=url)

    def set_http_proxy(self, host: str, port: int, username: str = None, password: str = None):
        url = f"http://{username}:{password}@{host}:{port}" if username and password else f"http://{host}:{port}"
        self._req.proxies = dict(http=url)

    def _get_main_session_captcha(self) -> Response:
        return self.get(f"{self.url}/main/sessionCaptcha")

    def _get_main_session_unlock_popup(self) -> Response:
        return self.get(f"{self.url}/main/sessionUnlockPopup")

    def _post_main_session_get_challenge(self, captcha_id: int, image_id: str = "") -> Response:
        c = [cookie.name for cookie in self._req.cookies if not cookie.has_nonstandard_attr("HttpOnly")]
        env = dict(l=["tets"], s=[], c=c, m=0)
        data = dict(_token=self.token, captchaId=captcha_id, env=utils.b64json(env))
        if image_id:
            data.update(imageId=image_id, isRefresh=True)
        return self.post(f"{self.url}/main/sessionGetChallenge", data=data)

    def _post_main_session_unlock(
        self, captcha_id: int, image_id: str, challenge_id: str, coords: List[Dict[str, int]], src: str
    ) -> Response:
        c = [cookie.name for cookie in self._req.cookies if not cookie.has_nonstandard_attr("HttpOnly")]
        env = dict(l=["tets"], s=[], c=c, m=0)
        cookies = dict(
            sh=hashlib.sha256(",".join(env["l"] + env["s"]).encode("utf8")).hexdigest(),
            ch=hashlib.sha256(",".join(env["c"]).encode("utf8")).hexdigest(),
        )
        cookie_kwargs = dict(
            expires=int(time.time()) + 120, path="/en/main/sessionUnlock", domain=".www.erepublik.com", secure=True, rest={"HttpOnly": True}
        )
        self._req.cookies.set("sh", cookies["sh"], **cookie_kwargs)
        self._req.cookies.set("ch", cookies["ch"], **cookie_kwargs)
        b64_env = utils.b64json(env)
        data = dict(
            _token=self.token,
            captchaId=captcha_id,
            imageId=image_id,
            challengeId=challenge_id,
            clickMatrix=utils.json_dumps(coords).replace(" ", ""),
            isMobile=0,
            env=b64_env,
            src=src,
        )
        return self.post(
            f"{self.url}/main/sessionUnlock",
            data=data,
            json=data,
            headers={"X-Requested-With": "XMLHttpRequest", "Referer": "https://www.erepublik.com/en"},
        )

    def _post_energy_refill_get_inventory(self):
        return self.post(f"{self.url}/economy/energyRefill-getInventory", data={"_token": self.token})


class ErepublikAnniversaryAPI(CitizenBaseAPI):
    def _post_main_collect_anniversary_reward(self) -> Response:
        return self.post(f"{self.url}/main/collect-anniversary-reward", data={"_token": self.token})

    # 12th anniversary endpoints
    def _get_anniversary_quest_data(self) -> Response:
        return self.get(f"{self.url}/main/anniversaryQuestData")

    def _post_map_rewards_unlock(self, node_id: int) -> Response:
        data = {"nodeId": node_id, "_token": self.token}
        return self.post(f"{self.url}/main/map-rewards-unlock", data=data)

    def _post_map_rewards_speedup(self, node_id: int, currency_amount: int) -> Response:
        data = {"nodeId": node_id, "_token": self.token, "currencyCost": currency_amount}
        return self.post(f"{self.url}/main/map-rewards-speedup", data=data)

    def _post_map_rewards_claim(self, node_id: int, extra: bool = False) -> Response:
        data = {"nodeId": node_id, "_token": self.token}
        if extra:
            data["claimExtra"] = 1
        return self.post(f"{self.url}/main/map-rewards-claim", data=data)

    def _post_main_wheel_of_fortune_spin(self, cost) -> Response:
        return self.post(f"{self.url}/main/wheeloffortune-spin", data={"_token": self.token, "_currentCost": cost})

    def _post_main_wheel_of_fortune_build(self) -> Response:
        return self.post(f"{self.url}/main/wheeloffortune-build", data={"_token": self.token})


class ErepublikArticleAPI(CitizenBaseAPI):
    def _get_main_article_json(self, article_id: int) -> Response:
        return self.get(f"{self.url}/main/articleJson/{article_id}")

    def _get_main_delete_article(self, article_id: int) -> Response:
        return self.get(f"{self.url}/main/delete-article/{article_id}/1")

    def _post_main_article_comments(self, article_id: int, page: int = 1) -> Response:
        data = dict(_token=self.token, articleId=article_id, page=page)
        if page:
            data.update({"page": page})
        return self.post(f"{self.url}/main/articleComments", data=data)

    def _post_main_article_comments_create(self, message: str, article_id: int, parent: int = 0) -> Response:
        data = dict(_token=self.token, message=message, articleId=article_id)
        if parent:
            data.update({"parentId": parent})
        return self.post(f"{self.url}/main/articleComments/create", data=data)

    def _post_main_donate_article(self, article_id: int, amount: int) -> Response:
        data = dict(_token=self.token, articleId=article_id, amount=amount)
        return self.post(f"{self.url}/main/donate-article", data=data)

    def _post_main_write_article(self, title: str, content: str, country_id: int, kind_id: int) -> Response:
        data = dict(_token=self.token, article_name=title, article_body=content, article_location=country_id, article_category=kind_id)
        return self.post(f"{self.url}/main/write-article", data=data)

    def _post_main_vote_article(self, article_id: int) -> Response:
        data = dict(_token=self.token, articleId=article_id)
        return self.post(f"{self.url}/main/vote-article", data=data)


class ErepublikCompanyAPI(CitizenBaseAPI):
    def _post_economy_assign_to_holding(self, factory_id: int, holding_id: int) -> Response:
        data = dict(_token=self.token, factoryId=factory_id, action="assign", holdingCompanyId=holding_id)
        return self.post(f"{self.url}/economy/assign-to-holding", data=data)

    def _post_economy_create_company(self, industry_id: int, building_type: int = 1) -> Response:
        data = {"_token": self.token, "company[industry_id]": industry_id, "company[building_type]": building_type}
        return self.post(f"{self.url}/economy/create-company", data=data, headers={"Referer": f"{self.url}/economy/create-company"})

    def _get_economy_inventory_items(self) -> Response:
        return self.get(f"{self.url}/economy/inventory-items/")

    def _get_economy_job_market_json(self, country_id: int) -> Response:
        return self.get(f"{self.url}/economy/job-market-json/{country_id}/1/desc")

    def _get_economy_my_companies(self) -> Response:
        return self.get(f"{self.url}/economy/myCompanies")

    def _post_economy_train(self, tg_ids: List[int]) -> Response:
        data: Dict[str, Union[int, str]] = {}
        for idx, tg_id in enumerate(tg_ids):
            data["grounds[%i][id]" % idx] = tg_id
            data["grounds[%i][train]" % idx] = 1
        if data:
            data["_token"] = self.token
        return self.post(f"{self.url}/economy/train", data=data)

    def _post_economy_upgrade_company(self, factory: int, level: int, pin: str = None) -> Response:
        data = dict(_token=self.token, type="upgrade", companyId=factory, level=level, pin="" if pin is None else pin)
        return self.post(f"{self.url}/economy/upgrade-company", data=data)

    def _post_economy_work(self, action_type: str, wam: List[int] = None, employ: Dict[int, int] = None) -> Response:
        data: Dict[str, Union[int, str]] = dict(action_type=action_type, _token=self.token)
        if action_type == "production":
            if employ is None:
                employ = {}
            if wam is None:
                wam = []
            max_idx = 0
            for company_id in sorted(wam or []):
                data.update(
                    {
                        f"companies[{max_idx}][id]": company_id,
                        f"companies[{max_idx}][employee_works]": employ.pop(company_id, 0),
                        f"companies[{max_idx}][own_work]": 1,
                    }
                )
                max_idx += 1
            for company_id in sorted(employ or []):
                data.update(
                    {
                        f"companies[{max_idx}][id]": company_id,
                        f"companies[{max_idx}][employee_works]": employ.pop(company_id, 0),
                        f"companies[{max_idx}][own_work]": 0,
                    }
                )
                max_idx += 1
        return self.post(f"{self.url}/economy/work", data=data)

    def _post_economy_work_overtime(self) -> Response:
        data = dict(action_type="workOvertime", _token=self.token)
        return self.post(f"{self.url}/economy/workOvertime", data=data)

    def _post_economy_job_market_apply(self, citizen_id: int, salary: float) -> Response:
        data = dict(_token=self.token, citizenId=citizen_id, salary=salary)
        return self.post(f"{self.url}/economy/job-market-apply", data=data)

    def _post_economy_resign(self) -> Response:
        return self.post(
            f"{self.url}/economy/resign",
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={"_token": self.token, "action_type": "resign"},
        )

    def _post_economy_sell_company(self, factory_id: int, pin: int = None, sell: bool = True) -> Response:
        data = dict(_token=self.token, pin="" if pin is None else pin)
        if sell:
            data.update({"sell": "sell"})
        else:
            data.update({"dissolve": factory_id})
        return self.post(f"{self.url}/economy/sell-company/{factory_id}", data=data, headers={"Referer": self.url})


class ErepublikCountryAPI(CitizenBaseAPI):
    def _get_country_military(self, country_name: str) -> Response:
        return self.get(f"{self.url}/country/military/{country_name}")

    def _post_main_country_donate(self, country_id: int, action: str, value: Union[int, float], quality: int = None) -> Response:
        data = dict(countryId=country_id, action=action, _token=self.token, value=value, quality=quality)
        return self.post(f"{self.url}/main/country-donate", data=data, headers={"Referer": f"{self.url}/country/economy/Latvia"})


class ErepublikEconomyAPI(CitizenBaseAPI):
    def _get_economy_citizen_accounts(self, organisation_id: int) -> Response:
        return self.get(f"{self.url}/economy/citizen-accounts/{organisation_id}")

    def _get_economy_my_market_offers(self) -> Response:
        return self.get(f"{self.url}/economy/myMarketOffers")

    def _get_main_job_data(self) -> Response:
        return self.get(f"{self.url}/main/job-data")

    def _post_main_buy_gold_items(self, currency: str, item: str, amount: int) -> Response:
        data = dict(itemId=item, currency=currency, amount=amount, _token=self.token)
        return self.post(f"{self.url}/main/buyGoldItems", data=data)

    def _post_economy_activate_booster(self, quality: int, duration: int, kind: str) -> Response:
        data = dict(type=kind, quality=quality, duration=duration, fromInventory=True, _token=self.token)
        return self.post(f"{self.url}/economy/activateBooster", data=data)

    def _post_economy_activate_house(self, quality: int) -> Response:
        data = dict(action="activate", quality=quality, type="house", _token=self.token)
        return self.post(f"{self.url}/economy/activateHouse", data=data)

    def _post_economy_donate_items_action(self, citizen_id: int, amount: int, industry: int, quality: int) -> Response:
        data = dict(citizen_id=citizen_id, amount=amount, industry_id=industry, quality=quality, _token=self.token)
        return self.post(
            f"{self.url}/economy/donate-items-action", data=data, headers={"Referer": f"{self.url}/economy/donate-items/{citizen_id}"}
        )

    def _post_economy_donate_money_action(self, citizen_id: int, amount: float = 0.0, currency: int = 62) -> Response:
        data = dict(citizen_id=citizen_id, _token=self.token, currency_id=currency, amount=amount)
        return self.post(
            f"{self.url}/economy/donate-money-action", data=data, headers={"Referer": f"{self.url}/economy/donate-money/{citizen_id}"}
        )

    def _post_economy_exchange_purchase(self, amount: float, currency: int, offer: int) -> Response:
        data = dict(_token=self.token, amount=amount, currencyId=currency, offerId=offer)
        return self.post(f"{self.url}/economy/exchange/purchase/", data=data)

    def _post_economy_exchange_retrieve(self, personal: bool, page: int, currency: int) -> Response:
        data = dict(_token=self.token, personalOffers=int(personal), page=page, currencyId=currency)
        return self.post(f"{self.url}/economy/exchange/retrieve/", data=data)

    def _post_economy_game_tokens_market(self, action: str) -> Response:
        assert action in [
            "retrieve",
        ]
        data = dict(_token=self.token, action=action)
        return self.post(f"{self.url}/economy/gameTokensMarketAjax", data=data)

    def _post_economy_marketplace(self, country: int, industry: int, quality: int, order_asc: bool = True) -> Response:
        data = dict(
            countryId=country,
            industryId=industry,
            quality=quality,
            ajaxMarket=1,
            orderBy="price_asc" if order_asc else "price_desc",
            _token=self.token,
        )
        return self.post(f"{self.url}/economy/marketplaceAjax", data=data)

    def _post_economy_marketplace_actions(self, action: str, **kwargs) -> Response:
        if action == "buy":
            data = dict(
                _token=self.token, offerId=kwargs["offer"], amount=kwargs["amount"], orderBy="price_asc", currentPage=1, buyAction=1
            )
        elif action == "sell":
            data = dict(
                _token=self.token,
                countryId=kwargs["country_id"],
                price=kwargs["price"],
                industryId=kwargs["industry"],
                quality=kwargs["quality"],
                amount=kwargs["amount"],
                sellAction="postOffer",
            )
        elif action == "delete":
            data = dict(_token=self.token, offerId=kwargs["offer_id"], sellAction="deleteOffer")
        else:
            raise ValueError(f"Action '{action}' is not supported! Only 'buy/sell/delete' actions are available")
        return self.post(f"{self.url}/economy/marketplaceActions", data=data)


class ErepublikLeaderBoardAPI(CitizenBaseAPI):
    def _get_main_leaderboards_damage_aircraft_rankings(self, country_id: int, weeks: int = 0, mu_id: int = 0) -> Response:  # noqa
        return self.get(f"{self.url}/main/leaderboards-damage-aircraft-rankings/{country_id}/{weeks}/{mu_id}/0")

    def _get_main_leaderboards_damage_rankings(self, country_id: int, weeks: int = 0, mu_id: int = 0, div: int = 0) -> Response:  # noqa
        return self.get(f"{self.url}/main/leaderboards-damage-rankings/{country_id}/{weeks}/{mu_id}/{div}")

    def _get_main_leaderboards_kills_aircraft_rankings(self, country_id: int, weeks: int = 0, mu_id: int = 0) -> Response:  # noqa
        return self.get(f"{self.url}/main/leaderboards-kills-aircraft-rankings/{country_id}/{weeks}/{mu_id}/0")

    def _get_main_leaderboards_kills_rankings(self, country_id: int, weeks: int = 0, mu_id: int = 0, div: int = 0) -> Response:  # noqa
        return self.get(f"{self.url}/main/leaderboards-kills-rankings/{country_id}/{weeks}/{mu_id}/{div}")


class ErepublikLocationAPI(CitizenBaseAPI):
    def _get_main_city_data_residents(self, city_id: int, page: int = 1, params: Mapping[str, Any] = None) -> Response:
        if params is None:
            params = {}
        return self.get(f"{self.url}/main/city-data/{city_id}/residents", params={"currentPage": page, **params})


class ErepublikMilitaryAPI(CitizenBaseAPI):
    def _get_military_battle_stats(self, battle_id: int, division: int, division_id: int):
        return self.get(f"{self.url}/military/battle-stats/{battle_id}/{division}/{division_id}")

    def _get_military_battlefield_choose_side(self, battle_id: int, side_id: int) -> Response:
        return self.get(f"{self.url}/military/battlefield-choose-side/{battle_id}/{side_id}")

    def _get_military_show_weapons(self, battle_id: int) -> Response:
        return self.get(f"{self.url}/military/show-weapons", params={"_token": self.token, "battleId": battle_id})

    def _get_military_campaigns(self) -> Response:
        return self.get(f"{self.url}/military/campaigns-new/")

    def _get_military_campaigns_json_list(self) -> Response:
        return self.get(f"{self.url}/military/campaignsJson/list")

    def _get_military_campaigns_json_citizen(self) -> Response:
        return self.get(f"{self.url}/military/campaignsJson/citizen")

    def _get_military_unit_data(self, unit_id: int, **kwargs) -> Response:
        params = {"groupId": unit_id, "panel": "members", **kwargs}
        return self.get(f"{self.url}/military/military-unit-data/", params=params)

    def _post_main_activate_battle_effect(self, battle_id: int, kind: str, citizen_id: int) -> Response:
        data = dict(battleId=battle_id, citizenId=citizen_id, type=kind, _token=self.token)
        return self.post(f"{self.url}/main/fight-activateBattleEffect", data=data)

    def _post_main_battlefield_travel(self, side_id: int, battle_id: int) -> Response:
        data = dict(_token=self.token, sideCountryId=side_id, battleId=battle_id)
        return self.post(f"{self.url}/main/battlefieldTravel", data=data)

    def _post_main_battlefield_change_division(self, battle_id: int, division_id: int, side_id: int = None) -> Response:
        data = dict(_token=self.token, battleZoneId=division_id, battleId=battle_id)
        if side_id is not None:
            data.update(sideCountryId=side_id)
        return self.post(f"{self.url}/main/battlefieldTravel", data=data)

    def _get_wars_show(self, war_id: int) -> Response:
        return self.get(f"{self.url}/wars/show/{war_id}")

    def _post_military_fight_activate_booster(self, battle_id: int, quality: int, duration: int, kind: str) -> Response:
        data = dict(type=kind, quality=quality, duration=duration, battleId=battle_id, _token=self.token)
        return self.post(f"{self.url}/military/fight-activateBooster", data=data)

    def _post_military_change_weapon(
        self,
        battle_id: int,
        battle_zone: int,
        weapon_level: int,
    ) -> Response:
        data = dict(battleId=battle_id, _token=self.token, battleZoneId=battle_zone, customizationLevel=weapon_level)
        return self.post(f"{self.url}/military/change-weapon", data=data)

    def _post_military_battle_console(self, battle_id: int, action: str, page: int = 1, **kwargs) -> Response:
        data = dict(battleId=battle_id, action=action, _token=self.token)
        if action == "battleStatistics":
            data.update(
                round=kwargs["round_id"],
                zoneId=kwargs["round_id"],
                leftPage=page,
                rightPage=page,
                division=kwargs["division"],
                type=kwargs.get("type", "damage"),
            )
        elif action == "warList":
            data.update(page=page)
        return self.post(f"{self.url}/military/battle-console", data=data)

    def _post_military_deploy_bomb(self, battle_id: int, division_id: int, side_id: int, bomb_id: int) -> Response:
        data = dict(battleId=battle_id, battleZoneId=division_id, sideId=side_id, sideCountryId=side_id, bombId=bomb_id, _token=self.token)
        return self.post(f"{self.url}/military/deploy-bomb", data=data)

    def _post_military_fight_air(self, battle_id: int, side_id: int, zone_id: int) -> Response:
        data = dict(sideId=side_id, battleId=battle_id, _token=self.token, battleZoneId=zone_id)
        return self.post(f"{self.url}/military/fight-shoooot/{battle_id}", data=data)

    def _post_military_fight_ground(self, battle_id: int, side_id: int, zone_id: int) -> Response:
        data = dict(sideId=side_id, battleId=battle_id, _token=self.token, battleZoneId=zone_id)
        return self.post(f"{self.url}/military/fight-shooot/{battle_id}", data=data)

    def _post_fight_deploy_deploy_report_data(self, deployment_id: int) -> Response:
        data = dict(_token=self.token, deploymentId=deployment_id)
        return self.post(f"{self.url}/military/fightDeploy-deployReportData", data=data)

    def _post_fight_deploy_get_inventory(self, battle_id: int, side_id: int, battle_zone_id: int) -> Response:
        data = dict(_token=self.token, battleId=battle_id, sideCountryId=side_id, battleZoneId=battle_zone_id)
        return self.post(f"{self.url}/military/fightDeploy-getInventory", data=data)

    def _post_fight_deploy_start_deploy(
        self, battle_id: int, side_id: int, battle_zone_id: int, energy: int, weapon: int, **kwargs
    ) -> Response:
        data = dict(
            _token=self.token,
            battleId=battle_id,
            battleZoneId=battle_zone_id,
            sideCountryId=side_id,
            weaponQuality=weapon,
            totalEnergy=energy,
            **kwargs,
        )
        return self.post(f"{self.url}/military/fightDeploy-startDeploy", data=data)

    def _post_military_fight_deploy_deploy_report_data(self, deployment_id: int) -> Response:
        data = dict(_token=self.token, deploymentId=deployment_id)
        return self.post(f"{self.url}/military/fightDeploy-deployReportData", data=data)


class ErepublikPoliticsAPI(CitizenBaseAPI):
    def _get_candidate_party(self, party_slug: str) -> Response:
        return self.get(f"{self.url}/candidate/{party_slug}")

    def _get_main_party_members(self, party_id: int) -> Response:
        return self.get(f"{self.url}/main/party-members/{party_id}")

    def _get_main_rankings_parties(self, country_id: int) -> Response:
        return self.get(f"{self.url}/main/rankings-parties/1/{country_id}")

    def _post_candidate_for_congress(self, presentation: str = "") -> Response:
        data = dict(_token=self.token, presentation=presentation)
        return self.post(f"{self.url}/candidate-for-congress", data=data)

    def _get_presidential_elections(self, country_id: int, timestamp: int) -> Response:
        return self.get(f"{self.url}/main/presidential-elections/{country_id}/{timestamp}")

    def _post_propose_president_candidate(self, party_slug: str, citizen_id: int) -> Response:
        return self.post(f"{self.url}/propose-president-candidate/{party_slug}", data=dict(_token=self.token, citizen=citizen_id))

    def _get_auto_propose_president_candidate(self, party_slug: str) -> Response:
        return self.get(f"{self.url}/auto-propose-president-candidate/{party_slug}")


class ErepublikPresidentAPI(CitizenBaseAPI):
    def _post_wars_attack_region(self, war_id: int, region_id: int, region_name: str) -> Response:
        data = {"_token": self.token, "warId": war_id, "regionName": region_name, "regionNameConfirm": region_name}
        return self.post(f"{self.url}/wars/attack-region/{war_id}/{region_id}", data=data)

    def _post_new_war(self, self_country_id: int, attack_country_id: int, debate: str = "") -> Response:
        data = dict(requirments=1, _token=self.token, debate=debate, countryNameConfirm=constants.COUNTRIES[attack_country_id].link)
        return self.post(f"{self.url}/{constants.COUNTRIES[self_country_id].link}/new-war", data=data)

    def _post_new_donation(self, country_id: int, amount: int, org_name: str, debate: str = "") -> Response:
        data = dict(requirments=1, _token=self.token, debate=debate, currency=1, value=amount, commit="Propose", type_name=org_name)
        return self.post(f"{self.url}/{constants.COUNTRIES[country_id].link}/new-donation", data=data)


class ErepublikProfileAPI(CitizenBaseAPI):
    def _get_main_citizen_hovercard(self, citizen_id: int) -> Response:
        return self.get(f"{self.url}/main/citizen-hovercard/{citizen_id}")

    def _get_main_citizen_profile_json(self, citizen_id: int) -> Response:
        return self.get(f"{self.url}/main/citizen-profile-json/{citizen_id}")

    def _get_main_citizen_notifications(self) -> Response:
        return self.get(f"{self.url}/main/citizenDailyAssistant")

    def _get_main_citizen_daily_assistant(self) -> Response:
        return self.get(f"{self.url}/main/citizenNotifications")

    def _get_main_messages_paginated(self, page: int = 1) -> Response:
        return self.get(f"{self.url}/main/messages-paginated/{page}")

    def _get_main_money_donation_accept(self, donation_id: int) -> Response:
        return self.get(f"{self.url}/main/money-donation/accept/{donation_id}", params={"_token": self.token})

    def _get_main_money_donation_reject(self, donation_id: int) -> Response:
        return self.get(f"{self.url}/main/money-donation/reject/{donation_id}", params={"_token": self.token})

    def _get_main_notifications_ajax_community(self, page: int = 1) -> Response:
        return self.get(f"{self.url}/main/notificationsAjax/community/{page}")

    def _get_main_notifications_ajax_system(self, page: int = 1) -> Response:
        return self.get(f"{self.url}/main/notificationsAjax/system/{page}")

    def _get_main_notifications_ajax_report(self, page: int = 1) -> Response:
        return self.get(f"{self.url}/main/notificationsAjax/report/{page}")

    def _get_main_training_grounds_json(self) -> Response:
        return self.get(f"{self.url}/main/training-grounds-json")

    def _get_main_weekly_challenge_data(self) -> Response:
        return self.get(f"{self.url}/main/weekly-challenge-data")

    def _post_main_citizen_add_remove_friend(self, citizen: int, add: bool) -> Response:
        data = dict(_token=self.token, citizenId=citizen, url="//www.erepublik.com/en/main/citizen-addRemoveFriend")
        if add:
            data.update({"action": "addFriend"})
        else:
            data.update({"action": "removeFriend"})
        return self.post(f"{self.url}/main/citizen-addRemoveFriend", data=data)

    def _post_main_daily_task_reward(self) -> Response:
        return self.post(f"{self.url}/main/daily-tasks-reward", data=dict(_token=self.token))

    def _post_delete_message(self, msg_id: list) -> Response:
        data = {"_token": self.token, "delete_message[]": msg_id}
        return self.post(f"{self.url}/main/messages-delete", data)

    def _post_eat(self, color: str) -> Response:
        data = dict(_token=self.token, buttonColor=color)
        return self.post(f"{self.url}/main/eat", params=data)

    def _post_main_global_alerts_close(self, alert_id: int) -> Response:
        data = dict(_token=self.token, alert_id=alert_id)
        return self.post(f"{self.url}/main/global-alerts/close", data=data)

    def _post_forgot_password(self, email: str) -> Response:
        data = dict(_token=self.token, email=email, commit="Reset password")
        return self.post(f"{self.url}/forgot-password", data=data)

    def _post_login(self, email: str, password: str) -> Response:
        data = dict(csrf_token=self.token, citizen_email=email, citizen_password=password, remember="on")
        return self.post(f"{self.url}/login", data=data)

    def _post_main_messages_alert(self, notification_ids: List[int]) -> Response:
        data = {"_token": self.token, "delete_alerts[]": notification_ids, "deleteAllAlerts": "1", "delete": "Delete"}
        return self.post(f"{self.url}/main/messages-alerts/1", data=data)

    def _post_main_notifications_ajax_community(self, notification_ids: List[int], page: int = 1) -> Response:
        data = {"_token": self.token, "delete_alerts[]": notification_ids}
        return self.post(f"{self.url}/main/notificationsAjax/community/{page}", data=data)

    def _post_main_notifications_ajax_system(self, notification_ids: List[int], page: int = 1) -> Response:
        data = {"_token": self.token, "delete_alerts[]": notification_ids}
        return self.post(f"{self.url}/main/notificationsAjax/system/{page}", data=data)

    def _post_main_notifications_ajax_report(self, notification_ids: List[int], page: int = 1) -> Response:
        data = {"_token": self.token, "delete_alerts[]": notification_ids}
        return self.post(f"{self.url}/main/notificationsAjax/report/{page}", data=data)

    def _post_main_messages_compose(self, subject: str, body: str, citizens: List[int]) -> Response:
        url_pk = 0 if len(citizens) > 1 else str(citizens[0])
        data = dict(citizen_name=",".join([str(x) for x in citizens]), citizen_subject=subject, _token=self.token, citizen_message=body)
        return self.post(f"{self.url}/main/messages-compose/{url_pk}", data=data)

    def _post_military_group_missions(self) -> Response:
        data = dict(action="check", _token=self.token)
        return self.post(f"{self.url}/military/group-missions", data=data)

    def _post_main_weekly_challenge_reward(self, reward_id: int) -> Response:
        data = dict(_token=self.token, rewardId=reward_id)
        return self.post(f"{self.url}/main/weekly-challenge-collect-reward", data=data)

    def _post_main_weekly_challenge_collect_all(self, max_reward_id: int) -> Response:
        data = dict(_token=self.token, maxRewardId=max_reward_id)
        return self.post(f"{self.url}/main/weekly-challenge-collect-all", data=data)

    def _post_main_profile_update(self, action: str, params: str):
        data = {"action": action, "params": params, "_token": self.token}
        return self.post(f"{self.url}/main/profile-update", data=data)


class ErepublikTravelAPI(CitizenBaseAPI):
    def _post_main_travel(self, check: str, **kwargs) -> Response:
        data = dict(_token=self.token, check=check, **kwargs)
        return self.post(f"{self.url}/main/travel", data=data)

    def _post_main_travel_data(self, **kwargs) -> Response:
        return self.post(f"{self.url}/main/travelData", data=dict(_token=self.token, **kwargs))


class ErepublikWallPostAPI(CitizenBaseAPI):
    # ## Country

    def _post_main_country_comment_retrieve(self, post_id: int) -> Response:
        data = {"_token": self.token, "postId": post_id}
        return self.post(f"{self.url}/main/country-comment/retrieve/json", data=data)

    def _post_main_country_comment_create(self, post_id: int, comment_message: str) -> Response:
        data = {"_token": self.token, "postId": post_id, "comment_message": comment_message}
        return self.post(f"{self.url}/main/country-comment/create/json", data=data)

    def _post_main_country_post_create(self, body: str, post_as: int) -> Response:
        data = {"_token": self.token, "post_message": body, "post_as": post_as}
        return self.post(f"{self.url}/main/country-post/create/json", data=data)

    def _post_main_country_post_retrieve(self) -> Response:
        data = {"_token": self.token, "page": 1, "switchedFrom": False}
        return self.post(f"{self.url}/main/country-post/retrieve/json", data=data)

    # ## Military Unit

    def _post_main_military_unit_comment_retrieve(self, post_id: int) -> Response:
        data = {"_token": self.token, "postId": post_id}
        return self.post(f"{self.url}/main/military-unit-comment/retrieve/json", data=data)

    def _post_main_military_unit_comment_create(self, post_id: int, comment_message: str) -> Response:
        data = {"_token": self.token, "postId": post_id, "comment_message": comment_message}
        return self.post(f"{self.url}/main/military-unit-comment/create/json", data=data)

    def _post_main_military_unit_post_create(self, body: str, post_as: int) -> Response:
        data = {"_token": self.token, "post_message": body, "post_as": post_as}
        return self.post(f"{self.url}/main/military-unit-post/create/json", data=data)

    def _post_main_military_unit_post_retrieve(self) -> Response:
        data = {"_token": self.token, "page": 1, "switchedFrom": False}
        return self.post(f"{self.url}/main/military-unit-post/retrieve/json", data=data)

    # ## Party

    def _post_main_party_comment_retrieve(self, post_id: int) -> Response:
        data = {"_token": self.token, "postId": post_id}
        return self.post(f"{self.url}/main/party-comment/retrieve/json", data=data)

    def _post_main_party_comment_create(self, post_id: int, comment_message: str) -> Response:
        data = {"_token": self.token, "postId": post_id, "comment_message": comment_message}
        return self.post(f"{self.url}/main/party-comment/create/json", data=data)

    def _post_main_party_post_create(self, body: str) -> Response:
        data = {"_token": self.token, "post_message": body}
        return self.post(f"{self.url}/main/party-post/create/json", data=data)

    def _post_main_party_post_retrieve(self) -> Response:
        data = {"_token": self.token, "page": 1, "switchedFrom": False}
        return self.post(f"{self.url}/main/party-post/retrieve/json", data=data)

    # ## Friend's Wall

    def _post_main_wall_comment_retrieve(self, post_id: int) -> Response:
        data = {"_token": self.token, "postId": post_id}
        return self.post(f"{self.url}/main/wall-comment/retrieve/json", data=data)

    def _post_main_wall_comment_create(self, post_id: int, comment_message: str) -> Response:
        data = {"_token": self.token, "postId": post_id, "comment_message": comment_message}
        return self.post(f"{self.url}/main/wall-comment/create/json", data=data)

    def _post_main_wall_post_create(self, body: str) -> Response:
        data = {"_token": self.token, "post_message": body}
        return self.post(f"{self.url}/main/wall-post/create/json", data=data)

    def _post_main_wall_post_retrieve(self) -> Response:
        data = {"_token": self.token, "page": 1, "switchedFrom": False}
        return self.post(f"{self.url}/main/wall-post/retrieve/json", data=data)

    # ## Medal posting
    def _post_main_wall_post_automatic(self, message: str, achievement_id: int) -> Response:
        return self.post(
            f"{self.url}/main/wall-post/automatic", data=dict(_token=self.token, message=message, achievementId=achievement_id)
        )


class CitizenAPI(
    ErepublikArticleAPI,
    ErepublikCountryAPI,
    ErepublikCompanyAPI,
    ErepublikEconomyAPI,
    ErepublikLeaderBoardAPI,
    ErepublikLocationAPI,
    ErepublikMilitaryAPI,
    ErepublikProfileAPI,
    ErepublikPresidentAPI,
    ErepublikPoliticsAPI,
    ErepublikAnniversaryAPI,
    ErepublikWallPostAPI,
    ErepublikTravelAPI,
):
    pass
